/*
 * Copyright (c) 2008  Yuri Chernyavsky
 * Copyright (c) 2007  Gregor Richards
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

/* TODO
 - testing
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <libgen.h>
#include <errno.h>
#include <assert.h>

#include "i18n.h"

#define SMV_HEADER_TITLE "SMV002000"

static unsigned int word2int(unsigned char word[3]) {
  return word[0] + (word[1] << 8) + (word[2] << 16);
}

static FILE *input;

/* There is non-null padding between the jpeg files.
   I do not know what it is for. I tried to replace the
   padding by '\0' bytes and the movie plays the same.
*/

typedef struct {
  unsigned char width[3];
  unsigned char height[3];
  unsigned char header_len[3];
  unsigned char const1[3];
  unsigned char jpeg_modulo[3];
  unsigned char fps[3];
  unsigned char num_frames[3];
  unsigned char const2[3];
  unsigned char const3[3];
  unsigned char frames_per_jpeg[3];
  unsigned char pad1[3];
  unsigned char pad2[3];
  unsigned char pad3[3];
  unsigned char pad4[3];
} smv_header_s;

static void print_info(smv_header_s *header) {
    int w, h, hl, c1, jm, fps, nf, c2, c3, fpj, p1, p2, p3, p4;   
 
    w = word2int(header->width);
    h = word2int(header->height);
    hl = word2int(header->header_len);
    c1 = word2int(header->const1);
    jm = word2int(header->jpeg_modulo);
    fps = word2int(header->fps);
    nf = word2int(header->num_frames);
    c2 = word2int(header->const2);
    c3 = word2int(header->const3);
    fpj = word2int(header->frames_per_jpeg);
    p1 = word2int(header->pad1);
    p2 = word2int(header->pad2);
    p3 = word2int(header->pad3);
    p4 = word2int(header->pad4);
    
    fprintf(stdout, 
"SMV file header:\n\
width           = %d\n\
height          = %d\n\
header_len      = %d\n\
const1          = %d\n\
jpeg_modulo     = %d\n\
frame_per_sec   = %d\n\
num_frames      = %d\n\
const2          = %d\n\
const3          = %d\n\
frames_per_jpeg = %d\n\
padding1        = %d\n\
padding2        = %d\n\
padding3        = %d\n\
padding4        = %d\n",
            w,h,hl,c1,jm,fps,nf,c2,c3,fpj,p1,p2,p3,p4);
}

/* Proceed audio */
static int proceed_wav(const char* wav_filename) {
  char *str, *ptr = strdup(SMV_HEADER_TITLE);
  FILE *wav_file;
  int ch;
  int count = 0;

  wav_file = fopen(wav_filename, "wb");

  str = ptr;
  /* seek for the smv header title */
  while (*str) {
    ch = fgetc(input);

    if (ch < 0 || ch == EOF) {
      fprintf(stdout, "unexpected error or EOF while reading audio\n");
      return -1;
    }

    if (ch != *str) {
      fwrite(ptr, 1, str - ptr, wav_file);
      fputc(ch, wav_file);
      count += (str - ptr + 1);  
      str = ptr;
    } else str++;
  }
  fclose(wav_file);
  return count;
}

/* Proceed the smv header */
static int proceed_header(smv_header_s* header) {
  if (fread(header, sizeof(smv_header_s), 1, input) != 1) { 
    fprintf(stdout, "failed to read smv header\n");
    return -1;
  }
  return 1;
}

static int proceed_jpeg(const char *filename, unsigned int jpeg_modulo) {
  unsigned char size[3];
  unsigned int n;
  unsigned int i;
  int ch, ret;
  FILE *file;
  char* err_str;

  ret = fread(size, 3, 1, input);
  if (ret != 1) {
    fprintf(stdout, "failed to read jpeg size for file=%s\n", filename);
    return -1;
  }

  n = word2int(size);

  file = fopen(filename, "wb");
  for (i=0; i<n; i++) {
    ch = fgetc(input);

    if (ch < 0 || ch == EOF) {
      fprintf(stdout, "unexpected error or EOF, file=%s\n", filename);
      return -1;
    }
    if (i == 0 && ch != 0xff) {
      fprintf(stdout, "bad jpeg start character, file=%s\n", filename);
      return -1;
    }
    if (i == n-1 && ch != 0xd9) {
      fprintf(stdout, "bad jpeg end character, file=%s\n", filename);
      return -1;
    }

    fputc(ch, file);
  }

  /* Read padding */  
  for (i=0; i<jpeg_modulo-n-3; i++) {
    ch = fgetc(input);
    if (ch < 0 || ch == EOF) {
      fprintf(stdout, "unexpected error or EOF in padding, file=%s\n", filename);
      return -1;
    }
  }

  fclose(file);
  return n;
}

int main(int argc, char **argv) {
  smv_header_s header;
  int jpeg_modulo;
  int num_jpeg;
  char *input_file_basename;
  char wav_file[2048];
  char output_dir[2048];
  char jpeg_dir[2048];
  char jpeg_file[2048];
  int i, n, ch, mj = 0;
  int ret;

  /* usage */
  if (argc != 2) {
      fprintf(stdout, "usage: %s <input_file>\n", argv[0]);
      exit(1);
  }

  input_file_basename = basename(strdup(argv[1]));
  ret = snprintf(output_dir, 1024, "%s.decode", input_file_basename);
  if (ret >= 1024) { 
     fprintf(stdout, "Input file name is too long\n");
     exit(-1);
  }

  input = fopen(argv[1], "rb");
  if (!input) {
      fprintf(stdout, "Cannot open input file\n");
      exit(-1);
  }

  sprintf(wav_file, "%s/audio.wav", output_dir);
  sprintf(jpeg_dir, "%s/frames", output_dir);

  mkdir(output_dir, S_IRWXU);
  mkdir(jpeg_dir, S_IRWXU);

  /* Proceed the wav file */
  n = proceed_wav(wav_file);

  if (ret <= 0) {
    exit(-1);
  }
  fprintf(stdout, "Audio: %d bytes\n", n);

  /* Proceed the smv header */
  if (proceed_header(&header) <=0) {
    exit(-1);
  }

  print_info(&header);
  jpeg_modulo = word2int(header.jpeg_modulo);
  num_jpeg = word2int(header.num_frames) / word2int(header.frames_per_jpeg);

  /* Proceed jpeg files */
  i = 0;
  for (;;) {
    if ((ch = fgetc(input)) == EOF) {
      if (i != num_jpeg) {
	fprintf(stdout, "read %d jpeg files, expected %d\n", i, num_jpeg);
      }
      break;
    }
      
    ungetc(ch, input);
    sprintf(jpeg_file, "%s/%010d.jpeg", jpeg_dir, i+1);
    n = proceed_jpeg(jpeg_file, jpeg_modulo);
    if (n <= 0) {
      exit(-1);
    }
    if (n > mj) mj = n;        

    i++;
  }
  fprintf(stdout, "Max jpeg size: %d\n", mj);
    
  return 0;
}
